from typing import Dict, List, Tuple

from tabulate import tabulate

import spyctl.api as api
import spyctl.config.configs as cfg
import spyctl.spyctl_lib as lib

SUMMARY_HEADERS = [
    "NAME",
    "SPYDERBAT_STATUS",
    "AGE",
    "VERSION",
    "CLUSTER",
    "MUID",
]


def nodes_output_summary(
    ctx: cfg.Context,
    clusters: List[str],
    time: Tuple[float, float],
    pipeline=None,
    limit_mem=False,
) -> str:
    data = []
    for node in api.get_nodes(
        *ctx.get_api_data(), clusters, time, pipeline, limit_mem
    ):
        data.append(__node_summary_data(node))
    rv = tabulate(
        sorted(data, key=lambda x: [x[4], x[1], x[2], x[0]]),
        headers=SUMMARY_HEADERS,
        tablefmt="plain",
    )
    return rv


def __node_summary_data(node: Dict) -> List[str]:
    k8s_status = node[lib.BE_K8S_STATUS]
    node_info = k8s_status[lib.NODE_INFO_FIELD]
    version = node_info.get(lib.KUBELET_VERSION_FIELD, lib.NOT_AVAILABLE)
    cluster = node.get("cluster_name") or node.get("cluster_uid")
    meta = node[lib.METADATA_FIELD]
    rv = [
        meta[lib.METADATA_NAME_FIELD],
        node["status"],
        lib.calc_age(lib.to_timestamp(meta[lib.METADATA_CREATE_TIME])),
        version,
        cluster,
        node.get("muid", lib.NOT_AVAILABLE),
    ]
    return rv



from typing import Dict, List, Tuple

from tabulate import tabulate

import spyctl.config.configs as cfg
import spyctl.spyctl_lib as lib
import spyctl.api as api
import spyctl.resources.api_filters as _af
import spyctl.filter_resource as filt

SUMMARY_HEADERS = ["NAME", "LAST_SEEN_STATUS", "AGE", "CLUSTER"]


def namespace_summary_output(
    name_or_uid: str,
    ctx: cfg.Context,
    clusters: List[str],
    time: Tuple[float, float],
    pipeline=None,
) -> str:
    data = []
    field_names = _af.Namespaces.get_name_or_uid_fields()
    for namespace in api.get_namespaces(
        *ctx.get_api_data(), clusters, time, pipeline
    ):
        ns = [namespace]
        if name_or_uid:
            ns = filt.filter_obj(ns, field_names, name_or_uid)
        if ns:
            data.append(__namespace_data(namespace))
    data.sort(key=lambda x: (x[3], x[0]))
    return tabulate(
        data,
        headers=SUMMARY_HEADERS,
        tablefmt="plain",
    )


def __namespace_data(namespace: Dict) -> List:
    meta = namespace[lib.METADATA_FIELD]
    rv = [
        meta[lib.METADATA_NAME_FIELD],
        "Active",
        lib.calc_age(lib.to_timestamp(meta[lib.METADATA_CREATE_TIME])),
        namespace["cluster_name"] or namespace["cluster_uid"],
    ]
    return rv



from typing import Dict, List, Tuple

import zulu
from tabulate import tabulate

import spyctl.api as api
import spyctl.config.configs as cfg
import spyctl.spyctl_lib as lib

NOT_AVAILABLE = lib.NOT_AVAILABLE
SUMMARY_HEADERS = [
    "NAME",
    "EXE",
    "ROOT_EXECUTION",
    "COUNT",
    "LATEST_EXECUTED",
]


class ProcessGroup:
    def __init__(self, multiple_exes) -> None:
        self.ref_proc = None
        self.latest_timestamp = NOT_AVAILABLE
        self.count = 0
        self.root = False
        self.multiple_exes = multiple_exes

    def add_proc(self, proc: Dict):
        self.__update_latest_timestamp(proc.get("create_time"))
        if self.ref_proc is None:
            self.ref_proc = proc
        if proc.get("euid") == 0:
            self.root = True
        self.count += 1

    def __update_latest_timestamp(self, timestamp):
        if timestamp is None:
            return
        if self.latest_timestamp == NOT_AVAILABLE:
            self.latest_timestamp = timestamp
        elif timestamp > self.latest_timestamp:
            self.latest_timestamp = timestamp

    def summary_data(self) -> List[str]:
        timestamp = NOT_AVAILABLE
        if self.latest_timestamp != NOT_AVAILABLE:
            timestamp = str(
                zulu.Zulu.fromtimestamp(self.latest_timestamp).format(
                    "YYYY-MM-ddTHH:mm:ss"
                )
            )
        exe = "MULTIPLE" if self.multiple_exes else self.ref_proc["exe"]
        rv = [
            self.ref_proc["name"],
            exe,
            "YES" if self.root else "NO",
            str(self.count),
            timestamp,
        ]
        return rv


def processes_stream_output_summary(
    ctx: cfg.Context,
    muids: List[str],
    time: Tuple[float, float],
    pipeline=None,
    limit_mem=False,
) -> str:
    groups: Dict[Tuple, ProcessGroup] = {}
    for proc in api.get_processes(
        *ctx.get_api_data(), muids, time, pipeline, limit_mem
    ):
        multiple_exes, key = _key(proc)
        if key not in groups:
            groups[key] = ProcessGroup(multiple_exes)
        groups[key].add_proc(proc)
    data = []
    for group in groups.values():
        data.append(group.summary_data())
    rv = tabulate(
        sorted(
            data,
            key=lambda x: [x[0], x[2], lib.to_timestamp(x[4])],
        ),
        headers=SUMMARY_HEADERS,
        tablefmt="plain",
    )
    return rv


def _key(process: Dict) -> Tuple:
    name = process["name"]
    exe = process["exe"]
    if exe.endswith(name):
        return False, (name, exe)
    return True, name



#
# cache_dict.py
#
# >>> import cache_dict
# >>> c = cache_dict.CacheDict(cache_len=2)
# >>> c[1] = 1
# >>> c[2] = 2
# >>> c[3] = 3
# >>> c
# CacheDict([(2, 2), (3, 3)])
# >>> c[2]
# 2
# >>> c[4] = 4
# >>> c
# CacheDict([(2, 2), (4, 4)])
# >>>
#

from collections import OrderedDict

# From https://gist.github.com/davesteele/44793cd0348f59f8fadd49d7799bd306


# -------------------------------------------------------------
class CacheDict(OrderedDict):
    """Dict with a limited length, ejecting LRUs as needed."""

    def __init__(self, *args, cache_len=None, on_del=None, **kwargs):
        self.cache_len = cache_len
        self.on_del = on_del
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        super().move_to_end(key)

        if self.cache_len is not None:
            while len(self) > self.cache_len:
                oldkey = next(iter(self))
                if self.on_del:
                    oldvalue = super().__getitem__(oldkey)
                    super().__delitem__(oldkey)
                    self.on_del(oldkey, oldvalue)
                else:
                    super().__delitem__(oldkey)

    def peek(self, key):
        return super().get(key)

    def get(self, key):
        if key in self:
            super().move_to_end(key)
        return super().get(key)

    def __getitem__(self, key):
        val = super().__getitem__(key)
        try:
            super().move_to_end(key)
        except KeyError:
            pass
        return val

    def contract(self, delta):
        self.cache_len -= delta
        self.cache_len = max(self.cache_len, 100)

    def expand(self, delta):
        self.cache_len += delta

    def flush(self):
        if self.on_del:
            for key, value in self.items():
                self.on_del(key, value)



import json
from typing import List, Union

import spyctl.schemas_v2 as schemas
from fastapi import APIRouter, BackgroundTasks
from pydantic import BaseModel, Field, Json
from typing_extensions import Annotated

import app.app_lib as app_lib
import app.commands.merge as cmd_merge

router = APIRouter(prefix="/api/v1")

PrimaryObject = Annotated[
    Union[schemas.GuardianBaselineModel, schemas.GuardianPolicyModel],
    Field(discriminator="kind"),
]

MergeObject = Annotated[
    Union[
        schemas.GuardianBaselineModel,
        schemas.GuardianDeviationModel,
        schemas.GuardianFingerprintGroupModel,
        schemas.GuardianFingerprintModel,
        schemas.GuardianPolicyModel,
        schemas.UidListModel,
    ],
    Field(discriminator="kind"),
]

# ------------------------------------------------------------------------------
# Merge Object(s) into Object
# ------------------------------------------------------------------------------


class MergeHandlerInput(BaseModel):
    object: Json[PrimaryObject] = Field(
        title="The primary object of the merge"
    )
    merge_objects: Json[List[MergeObject]] = Field(
        title="The object(s) to merge into the primary object."
    )
    org_uid: str
    api_key: str
    api_url: str


class MergeHandlerOutput(BaseModel):
    merged_object: str


@router.post("/merge")
def merge(
    i: MergeHandlerInput,
    background_tasks: BackgroundTasks,
) -> MergeHandlerOutput:
    background_tasks.add_task(app_lib.flush_spyctl_log_messages)
    merge_objects = [
        json.loads(obj.json(by_alias=True, exclude_unset=True))
        for obj in i.merge_objects
    ]
    cmd_input = cmd_merge.MergeInput(
        json.loads(i.object.json(by_alias=True, exclude_unset=True)),
        merge_objects,
        i.org_uid,
        i.api_key,
        i.api_url,
    )
    output = cmd_merge.merge(cmd_input)
    return MergeHandlerOutput(merged_object=output.merged_object)



import spyctl.cli as cli
import spyctl.spyctl_lib as lib
import spyctl.resources.suppression_policies as sp
import spyctl.search as search
import spyctl.commands.apply as apply


def handle_suppress_trace_by_id(orig_id: str, include_users: bool):
    id = orig_id
    trace = search.search_for_trace_by_uid(id)
    if not trace:
        exit(1)
    summary_uid = trace.get(lib.TRACE_SUMMARY_FIELD)
    if not summary_uid:
        cli.err_exit(f"Unable to find a Trace Summary for Trace {id}")
    t_sum = search.search_for_trace_summary_by_uid(summary_uid)
    if not t_sum:
        exit(1)
    pol = sp.build_trace_suppression_policy(t_sum, include_users)
    cli.try_log("")
    if not prompt_upload_policy(pol):
        cli.try_log("Operation cancelled.")
        return
    apply.handle_apply_suppression_policy(pol.as_dict())


def prompt_upload_policy(pol: sp.TraceSuppressionPolicy) -> bool:
    query = "Scope:\n-------------\n"
    query += pol.policy_scope_string
    query += "\nSuppress spydertraces within this scope?"
    return cli.query_yes_no(query)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Spyctl"
copyright = "2023, SPYDERBAT, Inc., All Rights Reserved"
author = "Spyderbat"
release = "0.10"
version = "0.11.1"


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx_copybutton",
    "sphinx_tabs.tabs",
    "sphinx_click",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
html_favicon = "_static/sb-logo.png"
html_logo = "_static/sb-logo-sidebar.png"



import json


# Mocking API Functionality
def mock_get_sources(api_url, api_key, org_uid):
    mock_source = {
        "uid": "mach:xxxXXxxxX",
        "name": "spyderbat-dev",
        "org_uid": "spyderbatuid",
        "runtime_description": "spyderbat-dev",
        "runtime_details": {
            "agent_registration_uid": "",
            "src_uid": "mach:-Y4VTygrRj8",
            "ip_addresses": [
                "172.17.0.1",
            ],
            "mac_addresses": [
                "ff:ff:ff:ff:ff:ff",
            ],
            "request_ip": "1.1.1.1",
            "hostname": "spyderbat-dev",
            "agent_arch": "x86_64",
            "agent_version": "v1.1.71",
            "uname": "5.4.0-153-generic",
            "os_name": "linux",
            "os_pretty_name": "Ubuntu 20.04.6 LTS",
            "boot_time": 1689379897.2830372,
            "cpu_make": "GenuineIntel",
            "cpu_model": "Intel(R) Core(TM) i9-9980HK CPU @ 2.40GHz",
            "cpu_cores": 4,
            "memory_total_gb": 15.622344970703125,
            "agent_status": "Agent Running",
            "agent_type": 0,
        },
        "valid_from": "2023-07-11T17:34:23Z",
        "valid_to": "0001-01-01T00:00:00Z",
        "resource_name": "srn:agent::xxxxxxxxx:xxxxxxxxxxxxxxxx",
        "last_data": "2023-07-15T12:17:44Z",
        "last_ingest_chunk_end_time": "2023-07-15T12:17:39Z",
        "last_stored_chunk_end_time": "2023-07-15T12:17:39Z",
        "agent_registration_uid": "xxxxxxxxxxxxxxxxxxxxx",
        "agent_type": 0,
    }
    return [mock_source]


def mock_get_clusters(api_url, api_key, org_uid):
    mock_cluster = {
        "uid": "clus:xxxxxxxxxxx",
        "name": "mock_cluster",
        "org_uid": "spyderbatuid",
        "cluster_details": {
            "cluster_uid": "xxxxxxxxxxxxxxxxxxxxxxxx",
            "cluster_name": "mock_cluster",
            "agent_uid": "mach:xxxxxxxxxxxxx",
            "src_uid": "xxxxxxxxxxxxxxxxxxxx",
            "cluid": "clus:xxxxxxxxxxx",
            "spyder_tags": {
                "CLUSTER_NAME": "mock_cluster",
                "baz": "bat",
                "foobar": "true",
            },
        },
        "valid_from": "2022-11-22T15:08:28Z",
        "valid_to": "0001-01-01T00:00:00Z",
        "resource_name": "srn:cluster::xxxxxxxxxxxx:clus:xxxxxxxxxxxxxx",
        "last_data": "2022-11-22T18:11:35Z",
    }
    return [mock_cluster]


def mock_get_deployments(
    api_url,
    api_key,
    org_uid,
    clusters,
    time,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar_on_first: bool = False,
):
    mock_deployment = {
        "schema": "model_k8s_deployment::1.0.0",
        "id": "deployment:xxxxxxxx:xxxxxxxxx",
        "status": "active",
        "version": 1694581060,
        "time": 1694584863.7841268,
        "valid_from": 1683042868.1558533,
        "expire_at": 1694588399.999999,
        "cluster_uid": "clus:xxxxxxxxxxx",
        "clusterid": "xxxxxxxxxxxxxxxxxxxxxxxxxxx",
        "cluster_name": "staging-k8s",
        "kind": "Deployment",
        "metadata": {
            "annotations": {
                "deployment.kubernetes.io/revision": "5",
                "meta.helm.sh/release-name": "test-svc",
                "meta.helm.sh/release-namespace": "default",
            },
            "creationTimestamp": "2023-04-25T18:19:40Z",
            "generation": 5,
            "labels": {
                "app.kubernetes.io/instance": "test-svc",
                "app.kubernetes.io/managed-by": "Helm",
                "app.kubernetes.io/name": "test-svc",
                "app.kubernetes.io/version": "1.16.0",
                "helm.sh/chart": "test-svc-0.1.0",
            },
            "name": "test-svc",
            "namespace": "default",
            "resourceVersion": "55651882",
            "uid": "xxxxxxx-xxx-xxxx-xxxx-xxxxxxxxxx",
        },
        "kuid": "xxxxxxx-xxx-xxxx-xxxx-xxxxxxxxxx",
        "spec": {
            "progressDeadlineSeconds": 600,
            "replicas": 1,
            "revisionHistoryLimit": 10,
            "selector": {
                "matchLabels": {
                    "app.kubernetes.io/instance": "test-svc",
                    "app.kubernetes.io/name": "test-svc",
                }
            },
            "strategy": {
                "rollingUpdate": {"maxSurge": "25%", "maxUnavailable": "25%"},
                "type": "RollingUpdate",
            },
            "template": {
                "metadata": {
                    "annotations": {
                        "kubectl.kubernetes.io/restartedAt": "2023-07-19T13:36:02Z"
                    },
                    "creationTimestamp": None,
                    "labels": {
                        "app.kubernetes.io/instance": "test-svc",
                        "app.kubernetes.io/name": "test-svc",
                    },
                },
                "spec": {
                    "containers": [
                        {
                            "command": ["/bin/bash", "-c", "test"],
                            "image": "test:latest",
                            "imagePullPolicy": "Always",
                            "livenessProbe": {
                                "exec": {"command": ["true"]},
                                "failureThreshold": 3,
                                "periodSeconds": 10,
                                "successThreshold": 1,
                                "timeoutSeconds": 1,
                            },
                            "name": "test",
                            "readinessProbe": {
                                "exec": {"command": ["true"]},
                                "failureThreshold": 3,
                                "periodSeconds": 10,
                                "successThreshold": 1,
                                "timeoutSeconds": 1,
                            },
                            "resources": {},
                            "securityContext": {},
                            "terminationMessagePath": "/dev/termination-log",
                            "terminationMessagePolicy": "File",
                        }
                    ],
                    "dnsPolicy": "ClusterFirst",
                    "restartPolicy": "Always",
                    "schedulerName": "default-scheduler",
                    "securityContext": {},
                    "serviceAccount": "test-svc",
                    "serviceAccountName": "test-svc",
                    "terminationGracePeriodSeconds": 30,
                    "tolerations": [
                        {
                            "effect": "NoSchedule",
                            "key": "dedicated",
                            "operator": "Equal",
                            "value": "testGroup",
                        }
                    ],
                },
            },
        },
        "k8s_status": {
            "availableReplicas": 1,
            "conditions": [
                {
                    "lastTransitionTime": "2023-07-24T21:44:12Z",
                    "lastUpdateTime": "2023-07-24T21:44:12Z",
                    "message": "Deployment has minimum availability.",
                    "reason": "MinimumReplicasAvailable",
                    "status": "True",
                    "type": "Available",
                },
                {
                    "lastTransitionTime": "2023-04-25T18:37:39Z",
                    "lastUpdateTime": "2023-09-11T15:55:56Z",
                    "message": 'ReplicaSet "test-svc" has successfully progressed.',
                    "reason": "NewReplicaSetAvailable",
                    "status": "True",
                    "type": "Progressing",
                },
            ],
            "observedGeneration": 17,
            "readyReplicas": 1,
            "replicas": 1,
            "updatedReplicas": 1,
        },
    }
    yield mock_deployment


def mock_get_namespaces(
    api_url,
    api_key,
    org_uid,
    clusters,
    time,
    pipeline=None,
    disable_pbar_on_first: bool = False,
):
    mock_namespace = {
        "apiVersion": "v1",
        "kind": "Namespace",
        "metadata": {
            "creationTimestamp": "2023-06-20T14:25:40Z",
            "labels": {"kubernetes.io/metadata.name": "test"},
            "name": "test",
            "resourceVersion": "65281414",
            "uid": "6c2956ff-239e-4068-937d-3390a6ed8575",
        },
        "spec": {"N/A": "N/A"},
        "status": {"phase": "Active"},
        "cluster_uid": "clus:PMx9HGEG_ZE",
        "cluster_name": "productiondemo",
    }
    yield mock_namespace


def mock_get_nodes(
    api_url,
    api_key,
    org_uid,
    clusters,
    time,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar_on_first: bool = False,
):
    mock_node = {
        "schema": "model_k8s_node::1.0.0",
        "id": "node:xxxxxxxxxxxxxxxxxxxxxx",
        "status": "active",
        "version": 1689425045,
        "time": 1689425066.5426638,
        "valid_from": 1689166499.9791923,
        "expire_at": 1689425999.999999,
        "cluster_uid": "clus:xxxxxxxxxxx",
        "clusterid": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
        "cluster_name": "mock_cluster",
        "kind": "Node",
        "metadata": {
            "annotations": {},
            "creationTimestamp": "2023-02-08T18:01:53Z",
            "labels": {
                "app": "mock",
            },
            "name": "mock_node",
            "resourceVersion": "45556759",
            "uid": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
        },
        "kuid": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
        "spec": {},
        "k8s_status": {
            "addresses": [],
            "allocatable": {},
            "capacity": {},
            "conditions": [],
            "daemonEndpoints": {},
            "images": [],
            "nodeInfo": {},
            "volumesAttached": [],
            "volumesInUse": [],
        },
        "muid": "mach:FomNj5G1TJc",
    }
    yield mock_node


def mock_get_pods(
    api_url,
    api_key,
    org_uid,
    clusters,
    time,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar_on_first: bool = False,
):
    mock_pod = {
        "schema": "model_k8s_pod::1.0.0",
        "id": "pod:xxxxxxxxxxx:xxxxxxxxxx",
        "status": "active",
        "version": 1689422366,
        "time": 1689422464.61505,
        "valid_from": 1689166501.0163338,
        "expire_at": 1689425999.999999,
        "cluster_uid": "clus:xxxxxxxxxxx",
        "clusterid": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
        "cluster_name": "mock_cluster",
        "kind": "Pod",
        "metadata": {
            "annotations": {},
            "creationTimestamp": "2023-06-30T15:29:49Z",
            "generateName": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
            "labels": {
                "app": "mock",
            },
            "name": "mock_pod",
            "namespace": "mock",
            "ownerReferences": [],
            "resourceVersion": "21281954",
            "uid": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
        },
        "kuid": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
        "spec": {
            "containers": [],
            "dnsPolicy": "ClusterFirst",
            "enableServiceLinks": True,
            "nodeName": "mock_node",
            "preemptionPolicy": "PreemptLowerPriority",
            "priority": 0,
            "restartPolicy": "Always",
            "schedulerName": "default-scheduler",
            "securityContext": {},
            "serviceAccount": "xxxxxxxxxxxxxxxxxxxxxxxxxxx",
            "serviceAccountName": "xxxxxxxxxxxxxxxxxxxxxxxxxxx",
            "terminationGracePeriodSeconds": 30,
            "tolerations": [],
            "volumes": [],
        },
        "k8s_status": {"phase": "Running"},
        "owner_uid": "replicaset:xxxxxxxxxxxxxxxxxxxxxx",
        "owner_kind": "ReplicaSet",
        "owner_name": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
        "deployment_uid": "deployment:xxxxxxxxxxxxxxxxxxxxxx",
        "deployment_name": "mock_deployment",
        "node_uid": "node:xxxxxxxxxxx:xxxxxxxxxx",
        "muid": "mach:xxxxxxxxxxx",
    }
    yield mock_pod


def mock_get_redflags(
    api_url,
    api_key,
    org_uid,
    muids,
    time,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar_on_first: bool = False,
):
    mock_redflag = {
        "id": "event_alert:xxxxxxxxxxx:xxxxxx:xxxxx:curl",
        "schema": "event_redflag:suspicious_command:1.1.0",
        "description": "SYSTEM as UID 100 ran curl with suspicious command /usr/bin/curl http://mock.mock",
        "ref": "proc:xxxxxxxxxxx:xxxxxx:xxxxx",
        "short_name": "command_curl",
        "class": [
            "redflag",
            "proc",
            "command",
            "high_severity",
            "suspicious",
            "curl",
        ],
        "severity": "high",
        "time": 1689424074.0437524,
        "routing": "customer",
        "version": 2,
        "linkback": "https://mock.com",
        "muid": "mach:xxxxxxxxxxx",
        "container": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
        "container_uid": "cont:xxxxxxxxxxapi_url, api_key, org_uid, timex:xxxxxxxxxxx:xxxxxxxxxxxx",
        "cluster_uid": "clus:xxxxxxxxxxx",
        "pod_uid": "pod:xxxxxxxxxxx:xxxxxxxxxx",
        "mitre_mapping": [
            {
                "technique": "T1105",
                "technique_name": "Ingress Tool Transfer",
                "url": "https://attack.mitre.org/techniques/T1105",
                "created": "2017-05-31T21:31:16.408Z",
                "modified": "2020-03-20T15:42:48.595Z",
                "stix": "attack-pattern--e6919abc-99f9-4c6c-95a5-14761e7b2add",
                "tactic": "TA0011",
                "tactic_name": "Command and Control",
                "platform": "Linux",
            }
        ],
        "auid": 4294967295,
        "name": "curl",
        "args": ["/usr/bin/curl", "http://mock.mock"],
        "auser": "SYSTEM",
        "euser": "UID 100",
        "ancestors": ["sh", "5", "runc", "containerd-shim", "systemd"],
        "false_positive": False,
        "uptime": 5498575.702408075,
        "traces": ["trace:xxxxxxxxxxx:xxxxxxxxxxx:xxxxx:suspicious_command"],
        "traces_suppressed": False,
    }
    yield mock_redflag


def mock_get_opsflags(
    api_url,
    api_key,
    org_uid,
    muids,
    time,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar_on_first: bool = False,
):
    mock_opsflag = {
        "id": "event_alert:xxxxxxxxxxx",
        "schema": "event_opsflag:memoryleak:1.1.0",
        "description": "Memory leaking over time",
        "ref": "mach:xxxxxxxxxxx",
        "short_name": "memory_leak",
        "class": ["opsflag", "mach", "memory_leak", "medium_severity"],
        "severity": "medium",
        "time": 1689422557.9508052,
        "routing": "customer",
        "version": 1,
        "linkback": "https://mock.mock",
        "muid": "mach:xxxxxxxxxxx",
    }
    yield mock_opsflag


def mock_get_fingerprints(
    api_url,
    api_key,
    org_uid,
    sources,
    time,
    fprint_type=None,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar_on_first: bool = False,
):
    mock_fingerprint = {
        "expire_at": 1689425999.999999,
        "id": "fprint:linux-service:xxxxxxxxxxx:update-motd.service:xxxxxx:xxxx",
        "cgroup": "systemd:/system.slice/update-motd.service",
        "muid": "mach:xxxxxxxxxxx",
        "apiVersion": "spyderbat/v1",
        "kind": "SpyderbatFingerprint",
        "metadata": {
            "name": "update-motd.service",
            "id": "fprint:linux-service:xxxxxxxxxxx:update-motd.service:xxxxxx:xxxx",
            "type": "linux-service",
            "checksum": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
            "org_uid": "spyderbatuid",
            "muid": "mach:xxxxxxxxxxx",
            "root": "proc:xxxxxxxxxxx:xxxxxx:9785",
            "firstTimestamp": 1689423122.9874985,
            "latestTimestamp": 1689423288.2522442,
            "version": 1689423128,
        },
        "spec": {
            "serviceSelector": {
                "cgroup": "systemd:/system.slice/update-motd.service"
            },
            "machineSelector": {"hostname": "mock_machine"},
            "processPolicy": [
                {
                    "name": "update-motd",
                    "exe": ["/usr/bin/bash"],
                    "id": "update-motd_0",
                    "euser": ["root"],
                }
            ],
            "networkPolicy": {
                "ingress": [],
                "egress": [
                    {
                        "to": [{"ipBlock": {"cidr": "1.1.1.1/32"}}],
                        "processes": ["yum_0"],
                        "ports": [{"protocol": "TCP", "port": 80}],
                    },
                ],
            },
        },
        "root_puid": "proc:xxxxxxxxxxx:xxxxxx:9785",
        "service_name": "update-motd.service",
        "proc_fprint_len": 18,
        "ingress_len": 0,
        "egress_len": 3,
        "schema": "model_fingerprint:linux_svc:1.0.0",
        "status": "closed",
        "time": 1689423288.2522442,
        "valid_from": 1689423122.9874985,
        "last_seen": 1689423126.8817828,
        "version": 1689423128,
        "checksum": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
        "valid_to": 1689423288.2522442,
    }
    yield mock_fingerprint


def mock_get_policies(api_url, api_key, org_uid, params=None):
    mock_policy = {
        "apiVersion": "spyderbat/v1",
        "kind": "SpyderbatPolicy",
        "metadata": {
            "name": "spyderbat-test",
            "type": "container",
            "latestTimestamp": 1672333396.3253918,
            "creationTimestamp": 1672333396.3253918,
            "uid": "1FZEoVkeS82aSI9jfLzm",
        },
        "spec": {
            "containerSelector": {
                "image": "spyderbat-test",
                "imageID": "sha256:6e2e1bce440ec41f53e849e56d5c6716ed7f1e1fa614d8dca2bbda49e5cde29e",
            },
            "podSelector": {
                "matchLabels": {
                    "app": "test",
                    "env": "prod",
                    "name": "test-web",
                    "pod-template-hash": "8665ffd6c6",
                    "tier": "frontend",
                }
            },
            "mode": "audit",
            "namespaceSelector": {
                "matchLabels": {"kubernetes.io/metadata.name": "test"}
            },
            "processPolicy": [
                {
                    "name": "python",
                    "exe": ["/usr/local/bin/python3.7"],
                    "id": "python_0",
                    "euser": ["root"],
                    "children": [
                        {
                            "name": "sh",
                            "exe": ["/bin/dash"],
                            "id": "sh_0",
                            "children": [
                                {
                                    "name": "uname",
                                    "exe": ["/bin/uname"],
                                    "id": "uname_0",
                                }
                            ],
                        }
                    ],
                },
                {
                    "name": "sh",
                    "exe": ["/bin/dash"],
                    "id": "sh_1",
                    "euser": ["root"],
                    "children": [
                        {
                            "name": "python",
                            "exe": ["/usr/local/bin/python3.7"],
                            "id": "python_1",
                            "euser": ["web-svc"],
                            "children": [
                                {
                                    "name": "sh",
                                    "exe": ["/bin/dash"],
                                    "id": "sh_2",
                                    "children": [
                                        {
                                            "name": "uname",
                                            "exe": ["/bin/uname"],
                                            "id": "uname_1",
                                        }
                                    ],
                                }
                            ],
                        }
                    ],
                },
            ],
            "networkPolicy": {
                "ingress": [
                    {
                        "from": [{"ipBlock": {"cidr": "192.168.6.11/32"}}],
                        "processes": ["python_0"],
                        "ports": [{"protocol": "TCP", "port": 22}],
                    },
                    {
                        "from": [{"ipBlock": {"cidr": "192.168.6.11/32"}}],
                        "processes": ["python_1"],
                        "ports": [{"protocol": "TCP", "port": 22}],
                    },
                ],
                "egress": [
                    {
                        "to": [{"dnsSelector": ["mongodb.local"]}],
                        "processes": ["python_0"],
                        "ports": [{"protocol": "TCP", "port": 27017}],
                    },
                    {
                        "to": [{"ipBlock": {"cidr": "192.168.5.10/32"}}],
                        "processes": ["python_0"],
                        "ports": [{"protocol": "TCP", "port": 443}],
                    },
                    {
                        "to": [{"dnsSelector": ["mongodb.local"]}],
                        "processes": ["python_1"],
                        "ports": [{"protocol": "TCP", "port": 27017}],
                    },
                    {
                        "to": [
                            {"ipBlock": {"cidr": "192.168.5.10/32"}},
                            {"ipBlock": {"cidr": "192.168.5.11/32"}},
                            {"ipBlock": {"cidr": "192.168.5.12/32"}},
                            {"ipBlock": {"cidr": "192.168.5.13/32"}},
                        ],
                        "processes": ["python_1"],
                        "ports": [{"protocol": "TCP", "port": 443}],
                    },
                ],
            },
            "response": {
                "default": [{"makeRedFlag": {"severity": "high"}}],
                "actions": [],
            },
        },
    }
    return [mock_policy]


def mock_post_new_policy(api_url, api_key, org_uid, data):
    class MockResponse:
        def __init__(self, json_data, status_code) -> None:
            self.status_code = status_code
            self.text = json.dumps(json_data)

    return MockResponse({"uid": "1FZEoVkeS82aSI9jfLzm"}, 200)


def mock_put_policy_update(api_url, api_key, org_uid, pol_uid, data):
    class MockResponse:
        def __init__(self, json_data, status_code) -> None:
            self.status_code = status_code
            self.text = json.dumps(json_data)

    return MockResponse({"uid": "1FZEoVkeS82aSI9jfLzm"}, 200)


def mock_delete_policy(api_url, api_key, org_uid, pol_uid):
    class MockResponse:
        def __init__(self, json_data, status_code) -> None:
            self.status_code = status_code
            self.text = json.dumps(json_data)

    return MockResponse({"uid": "1FZEoVkeS82aSI9jfLzm"}, 200)



from typing import IO, Dict, List, Set, Any

import ipaddress as ipaddr
import socket
from collections import defaultdict
import json
import tqdm
import sys

import spyctl.api as api
import spyctl.cli as cli
import spyctl.config.configs as cfg
import spyctl.filter_resource as filt
import spyctl.spyctl_lib as lib


def handle_describe(resource: str, name_or_uid: str, file: IO = None):
    if resource == lib.POLICIES_RESOURCE:
        handle_describe_policy(name_or_uid, file)


def handle_describe_policy(name_or_uid, file: IO = None):
    if not name_or_uid and not file:
        cli.err_exit("No file or name_or_uid provided")
    if not file:
        policy = get_policy_by_name_or_uid(name_or_uid)
    else:
        policy = get_policy_from_file(name_or_uid, file)
    pol_name = policy[lib.METADATA_FIELD][lib.METADATA_NAME_FIELD]
    cli.try_log(f"Describing policy '{pol_name}'")
    output = {
        "kind": policy[lib.KIND_FIELD],
        **policy[lib.METADATA_FIELD],
        "scope": describe_scope(policy[lib.SPEC_FIELD]),
    }
    output.update(describe_processes(policy[lib.SPEC_FIELD]))
    output.update(describe_connections(policy[lib.SPEC_FIELD]))
    cli.try_log("\n")
    cli.show(output, lib.OUTPUT_YAML)


def get_policy_from_file(name_or_uid: str, file: IO) -> Dict:
    resource = lib.load_resource_file(file)
    kind = resource[lib.KIND_FIELD]
    if kind != lib.POL_KIND:
        cli.err_exit(f"File kind '{kind}' is not '{lib.POL_KIND}'")
    if name_or_uid:
        policies = filt.filter_obj(
            [resource],
            [
                [lib.METADATA_FIELD, lib.METADATA_NAME_FIELD],
                [lib.METADATA_FIELD, lib.METADATA_UID_FIELD],
            ],
            name_or_uid,
        )
        if len(policies) == 0:
            cli.err_exit(f"No policies matching name_or_uid '{name_or_uid}'")
        resource = policies[0]
    type = resource[lib.METADATA_FIELD][lib.METADATA_TYPE_FIELD]
    if type not in lib.GUARDIAN_POL_TYPES:
        cli.err_exit(
            f"Policy type '{type}' is not a Guardian Policy type "
            f"'{lib.GUARDIAN_POL_TYPES}'"
        )
    return resource


def get_policy_by_name_or_uid(name_or_uid) -> Dict:
    ctx = cfg.get_current_context()
    policies = api.get_policies(*ctx.get_api_data())
    policies = filt.filter_obj(
        policies,
        [
            [lib.METADATA_FIELD, lib.METADATA_NAME_FIELD],
            [lib.METADATA_FIELD, lib.METADATA_UID_FIELD],
        ],
        name_or_uid,
    )
    if len(policies) == 0:
        cli.err_exit(f"No policies matching name_or_uid '{name_or_uid}'")
    if len(policies) > 1:
        cli.err_exit("Name or uid of policy is ambiguous")
    return policies[0]


def describe_scope(spec: Dict[str, Any]):
    rv = {}
    for key, value in spec.items():
        if key.endswith("Selector"):
            rv[key] = value
    return rv


def describe_processes(spec: Dict):
    proc_pol = spec[lib.PROC_POLICY_FIELD]
    proc_details = defaultdict(dict)
    users = set()
    node_count = [0]
    for proc in proc_pol:
        __get_proc_details(proc_details, users, node_count, proc)
    rv = {
        "Process Details": {
            "Total process nodes": node_count[0],
            "Effective users": list(users),
            "Processes": __sort_dict(__unset_dict(proc_details)),
        }
    }
    return rv


def __get_proc_details(
    proc_details: Dict[str, Dict],
    users: Set,
    node_count: List[int],
    proc: Dict,
):
    node_count[0] += 1
    name = proc[lib.NAME_FIELD]
    exes = proc[lib.EXE_FIELD]
    eusers = proc.get(lib.EUSER_FIELD, [])
    if name in proc_details:
        proc_details[name]["count"] += 1
        proc_details[name]["exes"].update(exes)
    else:
        proc_details[name]["count"] = 1
        proc_details[name]["exes"] = set(exes)
    users.update(eusers)
    children = proc.get(lib.CHILDREN_FIELD)
    if children:
        for child in children:
            __get_proc_details(proc_details, users, node_count, child)


def describe_connections(spec: Dict):
    net_pol = spec[lib.NET_POLICY_FIELD]
    ingress = net_pol[lib.INGRESS_FIELD]
    egress = net_pol[lib.EGRESS_FIELD]
    ips, hostnames, ports = extract_src_dst_info(ingress)
    i, h, p = extract_src_dst_info(egress)
    ips.update(i)
    hostnames.update(h)
    ports.update(p)
    dns_map, unknown_ips = __lookup_ips(ips)
    name_map, unknown_hostnames = __lookup_hostname(hostnames)
    for hostname, ip_set in name_map.items():
        if hostname in dns_map:
            dns_map[hostname].update(ip_set)
        else:
            dns_map[hostname] = ip_set
    rv = {
        "Connection Details": {
            "Total connection nodes": len(ingress) + len(egress),
            "Ports used": list(ports),
            "DNS MAP": {},
        }
    }
    rv["Connection Details"]["DNS MAP"].update(
        **__sort_dict(
            {h_name: list(ip_set) for h_name, ip_set in dns_map.items()}
        )
    )
    rv["Connection Details"]["DNS MAP"].update(
        **{"unknown ips": list(unknown_ips)}
    )
    rv["Connection Details"]["DNS MAP"].update(
        **{"unknown hostnames": list(unknown_hostnames)}
    )
    return rv


def extract_src_dst_info(ingress_or_egress):
    ips = set()
    hostnames = set()
    ports = set()
    for node in ingress_or_egress:
        block_list = node.get("from", node.get("to", {}))
        for block in block_list:
            if lib.IP_BLOCK_FIELD in block:
                ip = __eval_ip(block[lib.IP_BLOCK_FIELD][lib.CIDR_FIELD])
                if not ip:
                    continue
                ips.add(ip)
            elif lib.DNS_SELECTOR_FIELD in block:
                for name in block[lib.DNS_SELECTOR_FIELD]:
                    hostname = __eval_hostname(name)
                    if not hostname:
                        continue
                    hostnames.add(hostname)
        ports.update([p[lib.PORT_FIELD] for p in node[lib.PORTS_FIELD]])
    return ips, hostnames, ports


INCLUDE_PRIVATE = True


def __eval_ip(cidr: str):
    str_ip, bits = cidr.split("/")
    try:
        ip = ipaddr.IPv4Address(str_ip)
        if bits != "32":
            return None
        if ip.is_private and not INCLUDE_PRIVATE:
            return None
        return str_ip
    except ipaddr.AddressValueError:
        ip = ipaddr.IPv6Address(str_ip)
        if bits != "128":
            return None
        return str_ip


INCLUDE_LOCAL = True


def __eval_hostname(name: str):
    if not INCLUDE_LOCAL and name.endswith(".local"):
        return None
    return name


def __lookup_ips(ips: List[str]):
    if not ips:
        return {}, set()
    rv_map = defaultdict(set)
    unknown = set()
    cli.try_log("Looking up hostnames by IP")
    pbar = tqdm.tqdm(total=len(ips), leave=False, file=sys.stderr)
    for ip in ips:
        pbar.update(1)
        try:
            hostname, _, _ = socket.gethostbyaddr(ip)
        except Exception:
            unknown.add(ip)
            continue
        rv_map[hostname].add(ip)
    return rv_map, unknown


def __lookup_hostname(hostnames: List[str]):
    if not hostnames:
        return {}, set()
    rv_map = defaultdict(set)
    unknown = set()
    cli.try_log("Looking up IPs by hostname")
    pbar = tqdm.tqdm(total=len(hostnames), leave=False, file=sys.stderr)
    for hostname in hostnames:
        pbar.update(1)
        try:
            ip = socket.gethostbyname(hostname)
        except Exception:
            unknown.add(hostname)
            continue
        rv_map[hostname].add(ip)
    return rv_map, unknown


def __unset_dict(_dict: Dict) -> Dict:
    for key, value in _dict.items():
        if isinstance(value, set):
            _dict[key] = list(value)
        elif isinstance(value, dict):
            for v_key, v in value.items():
                if isinstance(v, set):
                    value[v_key] = list(v)
                elif isinstance(v, dict):
                    value[v_key] = __unset_dict(v)
    return _dict


def __sort_dict(_dict) -> Dict:
    dict_str = json.dumps(_dict, sort_keys=True)
    return json.loads(dict_str)



import json
import os

from fastapi.testclient import TestClient

from .main import app

client = TestClient(app)

API_KEY = os.environ.get("API_KEY", "test_key")
API_URL = os.environ.get("API_URL", "https://test.url.local")
ORG = os.environ.get("ORG", "test_org")

# cspell:disable
TEST_POLICY = '{\n  "apiVersion": "spyderbat/v1",\n  "kind": "SpyderbatPolicy",\n  "metadata": {\n    "name": "docker.io/guyduchatelet/spyderbat-demo:1",\n    "type": "container",\n    "latestTimestamp": 1686229419.97803\n  },\n  "spec": {\n    "containerSelector": {\n      "image": "docker.io/guyduchatelet/spyderbat-demo:1",\n      "imageID": "sha256:ce6595c9e9c3ed9faf8b7af095473d1d223c8cc4efa453c771767252d357c4af"\n    },\n    "podSelector": {\n      "matchLabels": {\n        "app": "rsvp",\n        "env": "prod",\n        "name": "rsvp-web",\n        "pod-template-hash": "5b4d5c9499",\n        "tier": "frontend"\n      }\n    },\n    "namespaceSelector": {\n      "matchLabels": {\n        "env": "prod",\n        "kubernetes.io/metadata.name": "rsvp-svc-prod"\n      }\n    },\n    "mode": "audit",\n    "processPolicy": [\n      {\n        "name": "python",\n        "exe": [\n          "/usr/local/bin/python3.7"\n        ],\n        "id": "python_0",\n        "euser": [\n          "root"\n        ],\n        "children": [\n          {\n            "name": "sh",\n            "exe": [\n              "/bin/dash"\n            ],\n            "id": "sh_0",\n            "children": [\n              {\n                "name": "uname",\n                "exe": [\n                  "/bin/uname"\n                ],\n                "id": "uname_0"\n              }\n            ]\n          }\n        ]\n      },\n      {\n        "name": "sh",\n        "exe": [\n          "/bin/dash"\n        ],\n        "id": "sh_1",\n        "euser": [\n          "root"\n        ],\n        "children": [\n          {\n            "name": "python",\n            "exe": [\n              "/usr/local/bin/python3.7"\n            ],\n            "id": "python_1",\n            "children": [\n              {\n                "name": "python",\n                "exe": [\n                  "/usr/local/bin/python3.7"\n                ],\n                "id": "python_2",\n                "children": [\n                  {\n                    "name": "sh",\n                    "exe": [\n                      "/bin/dash"\n                    ],\n                    "id": "sh_2",\n                    "children": [\n                      {\n                        "name": "uname",\n                        "exe": [\n                          "/bin/uname"\n                        ],\n                        "id": "uname_1"\n                      }\n                    ]\n                  }\n                ]\n              }\n            ]\n          }\n        ]\n      }\n    ],\n    "networkPolicy": {\n      "ingress": [],\n      "egress": [\n        {\n          "to": [\n            {\n              "dnsSelector": [\n                "mongodb.rsvp-svc-prod.svc.cluster.local"\n              ]\n            }\n          ],\n          "processes": [\n            "python_0",\n            "python_1",\n            "python_2"\n          ],\n          "ports": [\n            {\n              "protocol": "TCP",\n              "port": 27017\n            }\n          ]\n        }\n      ]\n    },\n    "response": {\n      "default": [\n        {\n          "makeRedFlag": {\n            "severity": "high"\n          }\n        }\n      ],\n      "actions": []\n    }\n  }\n}'  # noqa: E501
TEST_FINGERPRINT_LIST = '[\n  {\n    "apiVersion": "spyderbat/v1",\n    "kind": "SpyderbatFingerprint",\n    "metadata": {\n      "name": "docker.io/guyduchatelet/spyderbat-demo:1",\n      "id": "fprint:k8s-container:tTfnASZWRaY:ZIHFIg:559d01be326f",\n      "type": "container",\n      "checksum": "6faa5df4c3b452c4f13d77bdcb4021c9",\n      "org_uid": "KNUJF8M43WcT2o1qgEts",\n      "muid": "mach:tTfnASZWRaY",\n      "root": "proc:tTfnASZWRaY:ZIHFHw:2179798",\n      "containerID": "559d01be326f93d14f99fee190054a80e0f42f055be70ffd2d9e2f47c717fa4b",\n      "containerName": "docker.io/guyduchatelet/spyderbat-demo:1",\n      "firstTimestamp": 1686226210.5056648,\n      "latestTimestamp": 1686229419.97803,\n      "pod-uid": "pod:mupQ8S67xUo:vQi7hxaGf6",\n      "namespace": "rsvp-svc-prod",\n      "cluster-uid": "clus:mupQ8S67xUo",\n      "version": 1686229197\n    },\n    "spec": {\n      "containerSelector": {\n        "image": "docker.io/guyduchatelet/spyderbat-demo:1",\n        "imageID": "sha256:ce6595c9e9c3ed9faf8b7af095473d1d223c8cc4efa453c771767252d357c4af"\n      },\n      "podSelector": {\n        "matchLabels": {\n          "app": "rsvp",\n          "env": "prod",\n          "name": "rsvp-web",\n          "pod-template-hash": "5b4d5c9499",\n          "tier": "frontend"\n        }\n      },\n      "namespaceSelector": {\n        "matchLabels": {\n          "env": "prod",\n          "kubernetes.io/metadata.name": "rsvp-svc-prod"\n        }\n      },\n      "processPolicy": [\n        {\n          "name": "python",\n          "exe": [\n            "/usr/local/bin/python3.7"\n          ],\n          "id": "python_0",\n          "euser": [\n            "root"\n          ],\n          "children": [\n            {\n              "name": "sh",\n              "exe": [\n                "/bin/dash"\n              ],\n              "id": "sh_0",\n              "children": [\n                {\n                  "name": "uname",\n                  "exe": [\n                    "/bin/uname"\n                  ],\n                  "id": "uname_0"\n                }\n              ]\n            }\n          ]\n        },\n        {\n          "name": "sh",\n          "exe": [\n            "/bin/dash"\n          ],\n          "id": "sh_1",\n          "euser": [\n            "root"\n          ],\n          "children": [\n            {\n              "name": "python",\n              "exe": [\n                "/usr/local/bin/python3.7"\n              ],\n              "id": "python_1",\n              "children": [\n                {\n                  "name": "python",\n                  "exe": [\n                    "/usr/local/bin/python3.7"\n                  ],\n                  "id": "python_2",\n                  "children": [\n                    {\n                      "name": "sh",\n                      "exe": [\n                        "/bin/dash"\n                      ],\n                      "id": "sh_2"\n                    }\n                  ]\n                }\n              ]\n            }\n          ]\n        }\n      ],\n      "networkPolicy": {\n        "ingress": [],\n        "egress": [\n          {\n            "to": [\n              {\n                "dnsSelector": [\n                  "mongodb.rsvp-svc-prod.svc.cluster.local"\n                ]\n              }\n            ],\n            "processes": [\n              "python_0",\n              "python_1",\n              "python_2"\n            ],\n            "ports": [\n              {\n                "protocol": "TCP",\n                "port": 27017\n              }\n            ]\n          }\n        ]\n      }\n    }\n  },\n  {\n    "apiVersion": "spyderbat/v1",\n    "kind": "SpyderbatFingerprint",\n    "metadata": {\n      "name": "docker.io/guyduchatelet/spyderbat-demo:1",\n      "id": "fprint:k8s-container:tTfnASZWRaY:ZIHJ0g:6aa40d09d566",\n      "type": "container",\n      "checksum": "6faa5df4c3b452c4f13d77bdcb4021c9",\n      "org_uid": "KNUJF8M43WcT2o1qgEts",\n      "muid": "mach:tTfnASZWRaY",\n      "root": "proc:tTfnASZWRaY:ZIHJzg:2199996",\n      "containerID": "6aa40d09d566b64e5c230eb273ad946ee4df3fa147b8c70931333b25593ea64b",\n      "containerName": "docker.io/guyduchatelet/spyderbat-demo:1",\n      "firstTimestamp": 1686227410.6981585,\n      "latestTimestamp": 1686229419.97803,\n      "pod-uid": "pod:mupQ8S67xUo:FCOJMUldGw",\n      "namespace": "rsvp-svc-prod",\n      "cluster-uid": "clus:mupQ8S67xUo",\n      "version": 1686229197\n    },\n    "spec": {\n      "containerSelector": {\n        "image": "docker.io/guyduchatelet/spyderbat-demo:1",\n        "imageID": "sha256:ce6595c9e9c3ed9faf8b7af095473d1d223c8cc4efa453c771767252d357c4af"\n      },\n      "podSelector": {\n        "matchLabels": {\n          "app": "rsvp",\n          "env": "prod",\n          "name": "rsvp-web",\n          "pod-template-hash": "5b4d5c9499",\n          "tier": "frontend"\n        }\n      },\n      "namespaceSelector": {\n        "matchLabels": {\n          "env": "prod",\n          "kubernetes.io/metadata.name": "rsvp-svc-prod"\n        }\n      },\n      "processPolicy": [\n        {\n          "name": "python",\n          "exe": [\n            "/usr/local/bin/python3.7"\n          ],\n          "id": "python_0",\n          "euser": [\n            "root"\n          ],\n          "children": [\n            {\n              "name": "sh",\n              "exe": [\n                "/bin/dash"\n              ],\n              "id": "sh_0"\n            }\n          ]\n        },\n        {\n          "name": "sh",\n          "exe": [\n            "/bin/dash"\n          ],\n          "id": "sh_1",\n          "euser": [\n            "root"\n          ],\n          "children": [\n            {\n              "name": "python",\n              "exe": [\n                "/usr/local/bin/python3.7"\n              ],\n              "id": "python_1",\n              "children": [\n                {\n                  "name": "python",\n                  "exe": [\n                    "/usr/local/bin/python3.7"\n                  ],\n                  "id": "python_2",\n                  "children": [\n                    {\n                      "name": "sh",\n                      "exe": [\n                        "/bin/dash"\n                      ],\n                      "id": "sh_2",\n                      "children": [\n                        {\n                          "name": "uname",\n                          "exe": [\n                            "/bin/uname"\n                          ],\n                          "id": "uname_1"\n                        }\n                      ]\n                    }\n                  ]\n                }\n              ]\n            }\n          ]\n        }\n      ],\n      "networkPolicy": {\n        "ingress": [],\n        "egress": [\n          {\n            "to": [\n              {\n                "dnsSelector": [\n                  "mongodb.rsvp-svc-prod.svc.cluster.local"\n                ]\n              }\n            ],\n            "processes": [\n              "python_0",\n              "python_1",\n              "python_2"\n            ],\n            "ports": [\n              {\n                "protocol": "TCP",\n                "port": 27017\n              }\n            ]\n          }\n        ]\n      }\n    }\n  }\n]'  # noqa: E501
TEST_UID_LIST = '[{"apiVersion": "spyderbat/v1", "kind": "UidList", "metadata": {"startTime": 1690819253.8597443, "endTime": 1690820447.0023878}, "data": {"uniqueIdentifiers": ["fprint:k8s-container:IamnBjwlTR4:ZK7oaw:599d5eaa9314", "fprint:k8s-container:tTfnASZWRaY:ZK6kaQ:fed3ba83dde3"]}}]'  # noqa: E501
TEST_POLICY2 = '{"apiVersion": "spyderbat/v1", "kind": "SpyderbatPolicy", "metadata": {"name": "spyderbat-test", "type": "container", "latestTimestamp": 1672333396.3253918}, "spec": {"containerSelector": {"image": "spyderbat-test", "imageID": "sha256:6e2e1bce440ec41f53e849e56d5c6716ed7f1e1fa614d8dca2bbda49e5cde29e"}, "podSelector": {"matchLabels": {"app": "test", "env": "prod", "name": "test-web", "pod-template-hash": "8665ffd6c6", "tier": "frontend"}}, "namespaceSelector": {"matchLabels": {"kubernetes.io/metadata.name": "test"}}, "processPolicy": [{"name": "python", "exe": ["/usr/local/bin/python3.7"], "id": "python_0", "euser": ["root"], "children": [{"name": "sh", "exe": ["/bin/dash"], "id": "sh_0", "children": [{"name": "uname", "exe": ["/bin/uname"], "id": "uname_0"}]}]}, {"name": "sh", "exe": ["/bin/dash"], "id": "sh_1", "euser": ["root"], "children": [{"name": "python", "exe": ["/usr/local/bin/python3.7"], "id": "python_1", "euser": ["web-svc"], "children": [{"name": "sh", "exe": ["/bin/dash"], "id": "sh_2", "children": [{"name": "uname", "exe": ["/bin/uname"], "id": "uname_1"}]}]}]}], "networkPolicy": {"ingress": [{"from": [{"ipBlock": {"cidr": "192.168.6.11/32"}}], "processes": ["python_0", "python_1"], "ports": [{"protocol": "TCP", "port": 22}]}], "egress": [{"to": [{"dnsSelector": ["mongodb.local"]}], "processes": ["python_0", "python_1"], "ports": [{"protocol": "TCP", "port": 27017}]}, {"to": [{"ipBlock": {"cidr": "192.168.5.10/32"}}, {"ipBlock": {"cidr": "192.168.5.11/32"}}, {"ipBlock": {"cidr": "192.168.5.12/32"}}, {"ipBlock": {"cidr": "192.168.5.13/32"}}], "processes": ["python_0", "python_1"], "ports": [{"protocol": "TCP", "port": 443}]}]}, "response": {"default": [{"makeRedFlag": {"severity": "high"}}], "actions": []}}}'  # noqa: E501
TEST_POLICY3 = '{"apiVersion": "spyderbat/v1", "kind": "SpyderbatPolicy", "metadata": {"name": "spyderbat-demo-prod", "type": "container", "latestTimestamp": 1690465318.2703452, "uid": "hEzg8jyCCAGwu6OaG1KI", "creationTimestamp": "2023-07-27T13:46:21Z"}, "spec": {"containerSelector": {"image": "*spyderbat-demo*"}, "podSelector": {"matchLabels": {"app": "rsvp", "env": "prod", "name": "rsvp-web", "tier": "frontend"}}, "namespaceSelector": {"matchLabels": {"kubernetes.io/metadata.name": "rsvp-svc-prod"}}, "mode": "enforce", "processPolicy": [{"name": "sh", "exe": ["/bin/dash"], "id": "sh_0", "euser": ["root"], "children": [{"name": "python", "exe": ["/usr/local/bin/python3.7"], "id": "python_0", "children": [{"name": "python", "exe": ["/usr/local/bin/python3.7"], "id": "python_1", "children": [{"name": "sh", "exe": ["/bin/dash"], "id": "sh_2", "children": [{"name": "uname", "exe": ["/bin/uname"], "id": "uname_1"}]}]}, {"name": "sh", "exe": ["/bin/dash"], "id": "sh_1", "children": [{"name": "uname", "exe": ["/bin/uname"], "id": "uname_0"}]}]}]}, {"name": "python", "exe": ["/usr/local/bin/python3.7"], "id": "python_2", "euser": ["root"], "children": [{"name": "python", "exe": ["/usr/local/bin/python3.7"], "id": "python_3", "children": [{"name": "sh", "exe": ["/bin/dash"], "id": "sh_3", "children": [{"name": "uname", "exe": ["/bin/uname"], "id": "uname_3"}]}]}, {"name": "sh", "exe": ["/bin/dash"], "id": "sh_4", "children": [{"name": "uname", "exe": ["/bin/uname"], "id": "uname_2"}]}]}], "networkPolicy": {"ingress": [{"from": [{"ipBlock": {"cidr": "192.168.4.229/32"}}, {"ipBlock": {"cidr": "192.168.26.23/32"}}, {"ipBlock": {"cidr": "192.168.35.38/32"}}, {"ipBlock": {"cidr": "192.168.73.218/32"}}, {"ipBlock": {"cidr": "192.168.78.61/32"}}], "processes": ["python_1"], "ports": [{"protocol": "TCP", "port": 5000}]}, {"from": [{"ipBlock": {"cidr": "192.168.4.229/32"}}, {"ipBlock": {"cidr": "192.168.26.23/32"}}, {"ipBlock": {"cidr": "192.168.35.38/32"}}, {"ipBlock": {"cidr": "192.168.73.218/32"}}, {"ipBlock": {"cidr": "192.168.78.61/32"}}], "processes": ["python_3"], "ports": [{"protocol": "TCP", "port": 5000}]}], "egress": [{"to": [{"dnsSelector": ["mongodb.rsvp-svc-prod.svc.cluster.local"]}, {"ipBlock": {"cidr": "10.100.66.198/32"}}], "processes": ["python_0", "python_1"], "ports": [{"protocol": "TCP", "port": 27017}]}, {"to": [{"dnsSelector": ["mongodb.rsvp-svc-prod.svc.cluster.local"]}], "processes": ["python_3", "python_2"], "ports": [{"protocol": "TCP", "port": 27017}]}]}, "response": {"default": [{"makeRedFlag": {"severity": "high"}}], "actions": []}}}'  # noqa: E501
TEST_DEVIATIONS = '[{"apiVersion": "spyderbat/v1", "kind": "GuardianDeviation", "metadata": {"type": "ingress", "policy_uid": "hEzg8jyCCAGwu6OaG1KI"}, "spec": {"processPolicy": [{"policyNode": {"id": "python_1"}}], "networkPolicy": {"ingress": [{"from": [{"ipBlock": {"cidr": "192.168.47.160/32"}}], "ports": [{"protocol": "TCP", "port": 5000}], "processes": ["python_1"]}]}}}, {"apiVersion": "spyderbat/v1", "kind": "GuardianDeviation", "metadata": {"type": "ingress", "policy_uid": "hEzg8jyCCAGwu6OaG1KI"}, "spec": {"processPolicy": [{"policyNode": {"id": "python_1"}}], "networkPolicy": {"ingress": [{"from": [{"ipBlock": {"cidr": "192.168.31.148/32"}}], "ports": [{"protocol": "TCP", "port": 5000}], "processes": ["python_1"]}]}}}, {"apiVersion": "spyderbat/v1", "kind": "GuardianDeviation", "metadata": {"type": "ingress", "policy_uid": "hEzg8jyCCAGwu6OaG1KI"}, "spec": {"processPolicy": [{"policyNode": {"id": "python_1"}}], "networkPolicy": {"ingress": [{"from": [{"ipBlock": {"cidr": "192.168.33.128/32"}}], "ports": [{"protocol": "TCP", "port": 5000}], "processes": ["python_1"]}]}}}, {"apiVersion": "spyderbat/v1", "kind": "GuardianDeviation", "metadata": {"type": "ingress", "policy_uid": "hEzg8jyCCAGwu6OaG1KI"}, "spec": {"processPolicy": [{"policyNode": {"id": "python_1"}}], "networkPolicy": {"ingress": [{"from": [{"ipBlock": {"cidr": "192.168.78.0/24"}}], "ports": [{"protocol": "TCP", "port": 5000}], "processes": ["python_1"]}]}}}, {"apiVersion": "spyderbat/v1", "kind": "GuardianDeviation", "metadata": {"type": "ingress", "policy_uid": "hEzg8jyCCAGwu6OaG1KI"}, "spec": {"processPolicy": [{"policyNode": {"id": "python_1"}}], "networkPolicy": {"ingress": [{"from": [{"ipBlock": {"cidr": "192.168.95.203/32"}}], "ports": [{"protocol": "TCP", "port": 5000}], "processes": ["python_1"]}]}}}]'  # noqa: E501
# cspell:enable


def test_create_suppression_policy():
    data = {
        "type": "trace",
        "name": "Test Suppression Policy",
        "selectors": {
            "trigger-ancestors": ["systemd/foo/bar/baz"],
            "trigger-class": ["this/is/a/test/class"],
            "non-interactive-users": ["root"],
            "interactive_users": ["dev_*", "robert"],
        },
        "org_uid": ORG,
        "api_key": API_KEY,
        "api_url": API_URL,
    }
    print(data)
    response = client.post("/api/v1/create/suppressionpolicy", json=data)
    assert response.status_code == 200


def test_create_guardian_policy():
    data = {
        "input_objects": TEST_FINGERPRINT_LIST,
        "name": "Test Policy",
        "mode": "enforce",
        "org_uid": ORG,
        "api_key": API_KEY,
        "api_url": API_URL,
    }
    response = client.post("/api/v1/create/guardianpolicy", json=data)
    assert response.status_code == 200


def test_merge():
    data = {
        "object": TEST_POLICY,
        "merge_objects": TEST_FINGERPRINT_LIST,
        "org_uid": ORG,
        "api_key": API_KEY,
        "api_url": API_URL,
    }
    response = client.post("/api/v1/merge", json=data)
    assert response.status_code == 200


def test_diff():
    data = {
        "object": TEST_POLICY3,
        "diff_objects": TEST_DEVIATIONS,
        "org_uid": ORG,
        "api_key": API_KEY,
        "api_url": API_URL,
        "full_diff": True,
    }
    response = client.post("/api/v1/diff", json=data)
    data = {
        "object": TEST_POLICY3,
        "diff_objects": TEST_DEVIATIONS,
        "org_uid": ORG,
        "api_key": API_KEY,
        "api_url": API_URL,
        "content_type": "json",
    }
    response = client.post("/api/v1/diff", json=data)
    assert response.status_code == 200


def test_validate():
    data = {"object": TEST_POLICY}
    response = client.post("/api/v1/validate", json=data)
    assert response.status_code == 200
    invalid_message = response.json()["invalid_message"]
    assert not invalid_message
    data = {"object": json.dumps(TEST_POLICY2)}
    response = client.post("/api/v1/validate", json=data)
    assert response.status_code == 200
    invalid_message = response.json()["invalid_message"]
    assert not invalid_message


TEST_POLICY2 = {
    "apiVersion": "spyderbat/v1",
    "kind": "SpyderbatPolicy",
    "metadata": {
        "name": "spyderbat-test",
        "type": "container",
        "latestTimestamp": 1672333396.3253918,
    },
    "spec": {
        "containerSelector": {
            "image": "spyderbat-test",
            "imageID": "sha256:6e2e1bce440ec41f53e849e56d5c6716ed7f1e1fa614d8dca2bbda49e5cde29e",  # noqa: E501
        },
        "machineSelector": {
            "hostname": ["spyderbat-test"],
            "machineUID": "mach:12345678",
        },
        "mode": "audit",
        "processPolicy": [
            {
                "name": "python",
                "exe": ["/usr/local/bin/python3.7"],
                "id": "python_0",
                "euser": ["root"],
                "children": [
                    {
                        "name": "sh",
                        "exe": ["/bin/dash"],
                        "id": "sh_0",
                        "children": [
                            {
                                "name": "uname",
                                "exe": ["/bin/uname"],
                                "id": "uname_0",
                            }
                        ],
                    }
                ],
            },
            {
                "name": "sh",
                "exe": ["/bin/dash"],
                "id": "sh_1",
                "euser": ["root"],
                "children": [
                    {
                        "name": "python",
                        "exe": ["/usr/local/bin/python3.7"],
                        "id": "python_1",
                        "euser": ["web-svc"],
                        "children": [
                            {
                                "name": "sh",
                                "exe": ["/bin/dash"],
                                "id": "sh_2",
                                "children": [
                                    {
                                        "name": "uname",
                                        "exe": ["/bin/uname"],
                                        "id": "uname_1",
                                    }
                                ],
                            }
                        ],
                    }
                ],
            },
        ],
        "networkPolicy": {
            "ingress": [
                {
                    "from": [{"ipBlock": {"cidr": "192.168.6.11/32"}}],
                    "processes": ["python_0", "python_1"],
                    "ports": [{"protocol": "TCP", "port": 22}],
                }
            ],
            "egress": [
                {
                    "to": [{"dnsSelector": ["mongodb.local"]}],
                    "processes": ["python_0", "python_1"],
                    "ports": [{"protocol": "TCP", "port": 27017}],
                },
                {
                    "to": [
                        {"ipBlock": {"cidr": "192.168.5.10/32"}},
                        {"ipBlock": {"cidr": "192.168.5.11/32"}},
                        {"ipBlock": {"cidr": "192.168.5.12/32"}},
                        {"ipBlock": {"cidr": "192.168.5.13/32"}},
                    ],
                    "processes": ["python_0", "python_1"],
                    "ports": [{"protocol": "TCP", "port": 443}],
                },
            ],
        },
        "response": {
            "default": [{"makeRedFlag": {"severity": "high"}}],
            "actions": [],
        },
    },
}



from typing import IO, Dict, List, Optional, Union

import spyctl.api as api
import spyctl.cli as cli
import spyctl.config.configs as cfgs
import spyctl.filter_resource as filt
import spyctl.resources.policies as p
import spyctl.spyctl_lib as lib
import spyctl.commands.merge as merge_cmd
import spyctl.merge_lib as m_lib

POLICIES = None
FINGERPRINTS = None
MATCHING = "matching"
ALL = "all"
LOADED_RESOURCE = None


def handle_diff(
    filename_target: List[IO],
    policy_target: List[str],
    with_file: IO,
    with_policy: str,
    st,
    et,
    latest,
    merge_network=True,
    do_api=False,
    force_fprints=False,
    full_diff=False,
    output=lib.OUTPUT_DEFAULT,
):
    if do_api:
        ctx = cfgs.get_current_context()
        if not filename_target or not with_file:
            cli.err_exit("api test only with local files")
        r_data = lib.load_file_for_api_test(filename_target[0])
        w_data = lib.load_file_for_api_test(with_file)
        diff_data = api.api_diff(*ctx.get_api_data(), r_data, w_data)
        cli.show(diff_data, lib.OUTPUT_RAW)
        return
    global POLICIES
    if not POLICIES and (with_policy or policy_target):
        ctx = cfgs.get_current_context()
        POLICIES = api.get_policies(*ctx.get_api_data())
        POLICIES.sort(key=lambda x: x[lib.METADATA_FIELD][lib.NAME_FIELD])
        if not POLICIES and policy_target:
            cli.err_exit("No policies to diff.")
        elif not POLICIES and with_policy:
            cli.err_exit("No policies to diff with.")
    if filename_target:
        pager = True if len(filename_target) > 1 else False
        filename_target.sort(key=lambda x: x.name)
        for file in filename_target:
            target = load_target_file(file)
            target_name = f"local file '{file.name}'"
            resrc_kind = target.get(lib.KIND_FIELD)
            if resrc_kind not in [lib.BASELINE_KIND, lib.POL_KIND]:
                cli.try_log(
                    f"The 'diff' command is not supported for {resrc_kind}",
                    is_warning=True,
                )
                continue
            with_obj = get_with_obj(
                target,
                target_name,
                with_file,
                with_policy,
                st,
                et,
                latest,
                force_fprints,
            )
            if with_obj:
                diff_resource(
                    target,
                    target_name,
                    with_obj,
                    pager,
                    merge_network,
                    full_diff=full_diff,
                    output=output,
                )
            elif with_obj is False:
                continue
            else:
                cli.try_log(
                    f"{file.name} has nothing to diff with... skipping."
                )
                merge_obj = __nothing_to_diff_with(target_name, target, latest)
                if merge_obj:
                    handle_output(merge_obj, pager, full_diff=full_diff)
    elif policy_target:
        pager = (
            True
            if len(policy_target) > 1
            or (ALL in policy_target and len(POLICIES) > 1)
            else False
        )
        policy_target = sorted(policy_target)
        if ALL in policy_target:
            for target in POLICIES:
                t_name = lib.get_metadata_name(target)
                t_uid = target[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]
                target_name = f"applied policy '{t_name} - {t_uid}'"
                with_obj = get_with_obj(
                    target,
                    target_name,
                    with_file,
                    with_policy,
                    st,
                    et,
                    latest,
                    force_fprints,
                )
                if with_obj:
                    diff_resource(
                        target,
                        target_name,
                        with_obj,
                        pager,
                        merge_network,
                        full_diff=full_diff,
                        output=output,
                    )
                elif with_obj is False:
                    continue
                else:
                    merge_obj = __nothing_to_diff_with(
                        target_name, target, latest
                    )
                    if merge_obj:
                        handle_output(merge_obj, pager, full_diff=full_diff)
        else:
            targets = {}
            for pol_name_or_uid in policy_target:
                policies = filt.filter_obj(
                    POLICIES,
                    [
                        [lib.METADATA_FIELD, lib.NAME_FIELD],
                        [lib.METADATA_FIELD, lib.METADATA_UID_FIELD],
                    ],
                    pol_name_or_uid,
                )
                if len(policies) == 0:
                    cli.try_log(
                        "Unable to locate policy with name or UID"
                        f" {pol_name_or_uid}",
                        is_warning=True,
                    )
                    continue
                for policy in policies:
                    pol_uid = policy[lib.METADATA_FIELD][
                        lib.METADATA_UID_FIELD
                    ]
                    targets[pol_uid] = policy
            targets = sorted(
                list(targets.values()),
                key=lambda x: x[lib.METADATA_FIELD][lib.METADATA_NAME_FIELD],
            )
            if len(targets) > 1:
                pager = True
            for target in targets:
                t_name = lib.get_metadata_name(target)
                t_uid = target.get(lib.METADATA_FIELD, {}).get(
                    lib.METADATA_UID_FIELD
                )
                target_name = f"applied policy '{t_name} - {t_uid}'"
                with_obj = get_with_obj(
                    target,
                    target_name,
                    with_file,
                    with_policy,
                    st,
                    et,
                    latest,
                    force_fprints,
                )
                if with_obj:
                    diff_resource(
                        target,
                        target_name,
                        with_obj,
                        pager,
                        merge_network,
                        full_diff=full_diff,
                        output=output,
                    )
                elif with_obj is False:
                    continue
                else:
                    merge_obj = __nothing_to_diff_with(
                        target_name, target, latest
                    )
                    if merge_obj:
                        handle_output(merge_obj, pager, full_diff=full_diff)
    else:
        cli.err_exit("No target of the diff.")


def get_with_obj(
    target: Dict,
    target_name,
    with_file: IO,
    with_policy: str,
    st,
    et,
    latest,
    with_fingerprints=False,
) -> Optional[Union[Dict, List[Dict]]]:
    global FINGERPRINTS
    target_uid = target.get(lib.METADATA_FIELD, {}).get(lib.METADATA_UID_FIELD)
    if with_file:
        with_obj = load_with_file(with_file)
        if not cli.query_yes_no(
            f"diff {target_name} with '{with_file.name}'?"
        ):
            return False
    elif with_policy == MATCHING:
        if not target_uid:
            cli.try_log(
                f"{target_name} has no uid, unable to match with"
                " policy stored in the Spyderbat backend.",
                is_warning=True,
            )
            return False
        with_obj = get_with_policy(target_uid, POLICIES)
        if with_obj:
            pol_name = lib.get_metadata_name(with_obj)
            pol_uid = with_obj.get(lib.METADATA_FIELD, {}).get(
                lib.METADATA_UID_FIELD
            )
            if not cli.query_yes_no(
                f"diff {target_name} with data from applied policy"
                f" '{pol_name} - {pol_uid}'?"
            ):
                return False
    elif with_policy:
        with_obj = get_with_policy(with_policy, POLICIES)
        if with_obj:
            pol_name = lib.get_metadata_name(with_obj)
            pol_uid = with_obj.get(lib.METADATA_FIELD, {}).get(
                lib.METADATA_UID_FIELD
            )
            if not cli.query_yes_no(
                f"diff {target_name} with data from applied policy"
                f" '{pol_name} - {pol_uid}'?"
            ):
                return False
    elif with_fingerprints or not target.get(lib.METADATA_FIELD, {}).get(
        lib.METADATA_UID_FIELD
    ):
        if latest:
            latest_timestamp = target.get(lib.METADATA_FIELD, {}).get(
                lib.LATEST_TIMESTAMP_FIELD
            )
            if latest_timestamp is not None:
                st = lib.time_inp(latest_timestamp)
            else:
                cli.err_exit(
                    f"No {lib.LATEST_TIMESTAMP_FIELD} found in provided"
                    f" resource {lib.METADATA_FIELD} field. Defaulting to"
                    " 24hrs."
                )
            if FINGERPRINTS is not None:
                cli.try_log("--latest flag set, re-downloading fingerprints..")
                FINGERPRINTS = None
        if not cli.query_yes_no(
            f"diff {target_name} with relevant Fingerprints from"
            f" {lib.epoch_to_zulu(st)} to {lib.epoch_to_zulu(et)}?"
        ):
            return False
        if FINGERPRINTS is None:
            with_obj = get_with_fingerprints(target, st, et, latest)
            cli.try_log(f"Filtering fingerprints for {target_name}")
            with_obj = filter_fingerprints(target, with_obj)
        else:
            cli.try_log(f"Filtering fingerprints for {target_name}")
            with_obj = filter_fingerprints(target, FINGERPRINTS)
    else:
        if latest:
            st = merge_cmd.get_latest_timestamp(target)
        if not cli.query_yes_no(
            f"Diff {target_name} with Deviations from"
            f" {lib.epoch_to_zulu(st)} to {lib.epoch_to_zulu(et)}?"
        ):
            return False
        uid = target[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]
        with_obj = merge_cmd.get_with_deviations(uid, st, et)
    return with_obj


def load_target_file(target_file: IO) -> Dict:
    rv = lib.load_resource_file(target_file)
    return rv


def get_target_policy(target_uid) -> Optional[Dict]:
    rv = p.get_policy_by_uid(target_uid, POLICIES)
    filt.filter_policies()
    return rv


def load_with_file(with_file: IO) -> Dict:
    rv = lib.load_resource_file(with_file)
    return rv


def get_with_fingerprints(target: Dict, st, et, latest) -> List[Dict]:
    global FINGERPRINTS
    fingerprints = merge_cmd.get_with_fingerprints(target, st, et, latest)
    FINGERPRINTS = fingerprints
    return fingerprints


def filter_fingerprints(target, fingerprints) -> List[Dict]:
    filters = lib.selectors_to_filters(target)
    rv = filt.filter_fingerprints(
        fingerprints, **filters, use_context_filters=False
    )
    return rv


def get_with_policy(pol_uid: str, policies: List[Dict]) -> Dict:
    return p.get_policy_by_uid(pol_uid, policies)


def diff_resource(
    target: Dict,
    target_name,
    with_obj: Union[Dict, List[Dict]],
    pager=False,
    merge_network=True,
    full_diff=False,
    output=lib.OUTPUT_DEFAULT,
):
    merged_obj = merge_cmd.merge_resource(
        target, target_name, with_obj, "diff", merge_network
    )
    if merged_obj:
        handle_output(merged_obj, pager, full_diff, output)


def handle_output(
    merged_obj: m_lib.MergeObject,
    pager=False,
    full_diff=False,
    output=lib.OUTPUT_DEFAULT,
):
    if output == lib.OUTPUT_DEFAULT:
        output = lib.OUTPUT_RAW
        diff_object = False
    else:
        diff_object = True
    diff_data = merged_obj.get_diff(full_diff, diff_object)
    if pager:
        cli.show(diff_data, output, dest=lib.OUTPUT_DEST_PAGER)
    else:
        cli.show(diff_data, output)


def __nothing_to_diff_with(
    name: str, target, latest, src_cmd="diff"
) -> Optional[m_lib.MergeObject]:
    if latest:
        cli.try_log(
            f"{name.capitalize()} has nothing to {src_cmd} with. Would"
            f" update '{lib.LATEST_TIMESTAMP_FIELD}' field on merge."
        )
        return
        merge_object = merge_cmd.get_merge_object(
            target[lib.KIND_FIELD], target, True, src_cmd
        )
        merge_object.update_latest_timestamp()
        return merge_object
    else:
        cli.try_log(
            f"{name.capitalize()} has nothing to {src_cmd} with... skipping."
        )



import spyctl.cli as cli
import spyctl.schemas_v2 as schemas
import spyctl.spyctl_lib as lib


def handle_show_schema(resource: str):
    kind = __kind_helper(resource)
    if not kind:
        cli.err_exit(f"show-schema is not supported for '{kind}'")
    schema = schemas.handle_show_schema(kind)
    cli.show(schema, lib.OUTPUT_RAW)


def __kind_helper(resource: str):
    if resource == lib.BASELINES_RESOURCE:
        return lib.BASELINE_KIND
    if resource == lib.CONFIG_ALIAS:
        return lib.CONFIG_KIND
    if resource == lib.FINGERPRINTS_RESOURCE:
        return lib.FPRINT_KIND
    if resource == lib.FINGERPRINT_GROUP_RESOURCE:
        return lib.FPRINT_GROUP_KIND
    if resource == lib.POLICIES_RESOURCE:
        return lib.POL_KIND
    if resource == lib.SECRETS_ALIAS:
        return lib.SECRET_KIND
    if resource == lib.SUPPRESSION_POLICY_RESOURCE:
        return (lib.POL_KIND, lib.POL_TYPE_TRACE)
    if resource == lib.UID_LIST_RESOURCE:
        return lib.UID_LIST_KIND
    return None



import binascii
import json
from base64 import urlsafe_b64decode as b64url_d
from base64 import urlsafe_b64encode as b64url
from time import sleep
import time
from typing import Dict, List, Tuple

import spyctl.api as api
import spyctl.cli as cli
import spyctl.config.configs as cfg
import spyctl.filter_resource as filt
import spyctl.spyctl_lib as lib


def handle_logs(
    resource: str,
    name_or_id: str,
    follow: bool,
    st: float,
    et: float,
    tail: int,
    timestamps: bool,
    full: bool,
    since_iterator: str,
):
    if resource == lib.POLICIES_RESOURCE:
        handle_policy_logs(
            name_or_id, follow, st, et, tail, timestamps, full, since_iterator
        )
    else:
        cli.err_exit(f"The 'logs' command is not supported for '{resource}'")


def handle_policy_logs(
    name_or_uid: str,
    follow: bool,
    st: float,
    et: float,
    tail: int,
    timestamps: bool,
    full: bool,
    since_iterator: str,
):
    ctx = cfg.get_current_context()
    policies = api.get_policies(*ctx.get_api_data())
    policies = [
        p
        for p in filt.filter_obj(
            policies,
            [
                [lib.METADATA_FIELD, lib.METADATA_NAME_FIELD],
                [lib.METADATA_FIELD, lib.METADATA_UID_FIELD],
            ],
            name_or_uid,
        )
    ]
    if len(policies) == 0:
        cli.err_exit(f"No policies matching name_or_uid '{name_or_uid}'.")
    elif len(policies) > 1:
        cli.err_exit("Policy name is ambiguous, use uid.")
    policy = policies[0]
    policy_uid = policy[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]
    if not follow:
        if since_iterator:
            st, since_id = decode_audit_iterator(since_iterator)
            time = (st, et)
        else:
            time = (st, et)
            since_id = None
        audit_events = api.get_audit_events_tail(
            *ctx.get_api_data(), time, policy_uid, tail, since_id=since_id
        )
        show_policy_logs(audit_events, timestamps, full)
        if audit_events:
            e_time = audit_events[-1]["time"]
            e_id = audit_events[-1]["id"]
            iterator = encode_audit_iterator(e_time, e_id)
        else:
            iterator = "N/A"
        cli.try_log(f"Last Log Iterator: {iterator}")
    else:
        get_audit_events_follow(
            st,
            policy_uid,
            timestamps,
            full,
            tail,
            since_iterator=since_iterator,
        )


def get_audit_events_follow(
    st,
    src_uid,
    timestamps: bool,
    full: bool,
    tail: int = -1,
    msg_type=None,
    since_iterator=None,
) -> Tuple[List[Dict], str]:
    ctx = cfg.get_current_context()
    while True:
        et = time.time()
        if since_iterator:
            st, since_id = decode_audit_iterator(since_iterator)
            q_time = (st, et)
        else:
            q_time = (st, et)
            since_id = None
        audit_events = api.get_audit_events_tail(
            *ctx.get_api_data(),
            q_time,
            src_uid,
            tail,
            msg_type,
            since_id=since_id,
            disable_pbar=True,
        )
        tail = -1
        show_policy_logs(audit_events, timestamps, full)
        if audit_events:
            e_time = audit_events[-1]["time"]
            e_id = audit_events[-1]["id"]
            since_iterator = encode_audit_iterator(e_time, e_id)
        sleep(2.5)


def show_policy_logs(
    policy_audit_events: List[Dict], timestamps: bool, full: bool
):
    for event in policy_audit_events:
        if timestamps:
            ts = lib.epoch_to_zulu(event["time"]) + " -- "
        else:
            ts = ""
        if full:
            log = json.dumps(event)
        else:
            log = event["description"]
        msg = f"{ts}{log}"
        cli.show(msg, lib.OUTPUT_RAW)


def encode_audit_iterator(time: float, id: str):
    time -= 60
    iterator_bytes = f"{time}|{id}".encode("ascii")
    iterator = b64url(iterator_bytes).decode("ascii").strip("=")
    return iterator


def decode_audit_iterator(iterator: str):
    try:
        try:
            b = b64url_d(iterator + "==")
        except binascii.Error:
            try:
                b = b64url_d(iterator + "=")
            except binascii.Error:
                b = b64url_d(iterator)
        iterator_str = b.decode("ascii")
        time, id = iterator_str.split("|")
    except Exception:
        cli.err_exit("Unable to parse iterator")
    return float(time), id
